use std::sync::{Arc, Mutex};

use crate::{lox_context, DbConfig};
use chrono::prelude::*;
use lox_library::{BridgeAuth, BridgeDb};
use sled::IVec;

pub fn write_context_to_db(db: sled::Db, context: lox_context::LoxServerContext) {
    let date = Local::now().format("%Y-%m-%d_%H:%M:%S").to_string();
    let json_date = serde_json::to_vec(&date).unwrap();
    let json_result = serde_json::to_vec(&context).unwrap();
    let _ = db.insert(IVec::from(json_date), IVec::from(json_result));
}

pub fn find_existing_db(
    db_config: DbConfig,
) -> Result<(sled::Db, lox_context::LoxServerContext), sled::Error> {
    let context: lox_context::LoxServerContext;
    let (lox_db, context) = match sled::open(db_config.db_path) {
        Ok(lox_db) => {
            // Check if the lox_db already exists
            if lox_db.was_recovered() {
                // Check if there is a roll back date and try to choose the appropriate context
                // to rollback to, otherwise, take the last saved context
                match db_config.roll_back_date {
                    // If roll back date has been specified, either the exact date or range should be set
                    Some(roll_back_date) => {
                        // If the exact date is specified and it's in the database, use that to populate the context
                        match roll_back_date.exact_date {
                            Some(exact_date) => {
                                match lox_db.contains_key(exact_date.clone()) {
                                    // Find exact date/time in db and use the context from that date.
                                    Ok(_) => {
                                        let ivec_context = lox_db.get(exact_date).unwrap().unwrap();
                                        context = serde_json::from_slice(&ivec_context).unwrap();
                                    }
                                    // If the entered date is not found, use the last saved context
                                    Err(_) => {
                                        println!("UNEXPECTED DATE: Specified exact date not found in lox db! Using last saved context");
                                        context = use_last_context(lox_db.clone());
                                    }
                                }
                            }
                            // Otherwise check that the start of a range has been specified
                            None => {
                                match roll_back_date.date_range_start {
                                    Some(start) => {
                                        // If so, ensure the end of the range has also been specified
                                        if let Some(end) = roll_back_date.date_range_end {
                                            let start_json: Vec<u8> =
                                                serde_json::to_vec(&start).unwrap();
                                            let end_json: Vec<u8> =
                                                serde_json::to_vec(&end).unwrap();
                                            let r: sled::Iter = lox_db.range(
                                                IVec::from(start_json)..IVec::from(end_json),
                                            );
                                            let ivec_context = r.last().unwrap().unwrap().1;
                                            context =
                                                serde_json::from_slice(&ivec_context).unwrap();
                                        } else {
                                        println!("UNEXPECTED DATE: Start of range was specified but End of range was not! Using last saved context");
                                            context = use_last_context(lox_db.clone());
                                        }
                                    }
                                    // If not, roll_back_date should not have been indicated but it was! So just use the last
                                    // context and print that this is happening.
                                    None => {
                                        // If exact date and rollback range are blank, use the last saved context
                                        println!("UNEXPECTED DATE: No date specified despite indicating rollback! Using last saved context");
                                        context = use_last_context(lox_db.clone());
                                    }
                                }
                            }
                        }
                    }
                    // Use the last entry to populate the Lox context if no rollback date is set (which should be most common)
                    None => {
                        context = use_last_context(lox_db.clone());
                    }
                }
            //Otherwise, create a new Lox context
            } else {
                let new_db = BridgeDb::new();
                let new_ba = BridgeAuth::new(new_db.pubkey);
                context = lox_context::LoxServerContext {
                    db: Arc::new(Mutex::new(new_db)),
                    ba: Arc::new(Mutex::new(new_ba)),
                    extra_bridges: Arc::new(Mutex::new(Vec::new())),
                    to_be_replaced_bridges: Arc::new(Mutex::new(Vec::new())),
                }
            }
            (lox_db, context)
        }
        Err(e) => {
            panic!("Unable to read or create lox database! {:?}", e);
        }
    };
    Ok((lox_db, context))
}

fn use_last_context(cloned_db: sled::Db) -> lox_context::LoxServerContext {
    let ivec_context = cloned_db.last().unwrap().unwrap().1;
    println!("Using db date: {:?}", ivec_context);
    serde_json::from_slice(&ivec_context).unwrap()
}
